/*
 * Copyright 2015 - 2023 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

#pragma once

#include "array.hpp"
#include "bstring.hpp"
#include "random.hpp"
#include "stringview.hpp"
#include "strops.hpp"
#include "trace.hpp"

namespace ox {

using UUIDStr = ox::BString<36>;

namespace detail {

[[nodiscard]]
constexpr auto isHexChar(char c) noexcept {
	return (c >= '0' && c <= '9')
	    || (c >= 'a' && c <= 'f')
	    || (c >= 'A' && c <= 'F');
}

constexpr ox::Result<uint8_t> fromHex(ox::CRStringView v) noexcept {
	constexpr auto valMap = [] {
		ox::Array<uint8_t, 128> out;
		out['A'] = out['a'] = 10;
		out['B'] = out['b'] = 11;
		out['C'] = out['c'] = 12;
		out['D'] = out['d'] = 13;
		out['E'] = out['e'] = 14;
		out['F'] = out['f'] = 15;
		out['0'] = 0;
		out['1'] = 1;
		out['2'] = 2;
		out['3'] = 3;
		out['4'] = 4;
		out['5'] = 5;
		out['6'] = 6;
		out['7'] = 7;
		out['8'] = 8;
		out['9'] = 9;
		return out;
	}();
	if (!detail::isHexChar(v[0]) || !detail::isHexChar(v[1])) {
		return OxError(1, "Invalid UUID");
	}
	if (v.len() != 2) {
		return OxError(2);
	}
	uint8_t out = 0;
	out += valMap[static_cast<unsigned>(v[0])] * 16u;
	out += valMap[static_cast<unsigned>(v[1])];
	return out;
}

constexpr ox::BString<2> toHex(uint8_t v) noexcept {
	constexpr ox::Array<char, 16> valMap {
		'0',
		'1',
		'2',
		'3',
		'4',
		'5',
		'6',
		'7',
		'8',
		'9',
		'a',
		'b',
		'c',
		'd',
		'e',
		'f',
	};
	ox::Array<char, 3> out;
	out[0] = valMap[static_cast<unsigned>((v & 0xf0) / 0xf)];
	out[1] = valMap[static_cast<unsigned>(v & 0x0f)];
	out[2] = 0;
	return out.data();
}

}

class UUID {
	private:
		static bool s_seeded;
		static Random s_rand;
		ox::Array<uint8_t, 16> m_value;

	public:
		static void seedGenerator(const RandomSeed &seed) noexcept;

		static ox::Result<UUID> generate() noexcept;

		constexpr ox::Error fromString(ox::CRStringView s) noexcept {
			if (s.len() < 36) {
				return OxError(1, "Insufficient data contain complete UUID");
			}
			auto valueI = 0u;
			for (auto i = 0u; i < s.len();) {
				if (s[i] == '-') {
					++i;
					continue;
				}
				const auto seg = s.substr(i, i + 2);
				if (seg.len() != 2) {
					return OxError(1, "Invalid UUID");
				}
				oxRequire(val, detail::fromHex(seg));
				m_value[valueI] = val;
				i += 2;
				++valueI;
			}
			return {};
		}

		[[nodiscard]]
		constexpr UUIDStr toString() const noexcept {
			UUIDStr out;
			auto i = 0u;
			constexpr auto printChars = [](
					ox::BString<36> *out, const Array<uint8_t, 16> &value, std::size_t cnt, unsigned i) {
				for (; i < cnt; ++i) {
					const auto v = value[i];
					const auto h = detail::toHex(v);
					oxIgnoreError(out->append(h.c_str(), h.len()));
				}
				return i;
			};
			i = printChars(&out, m_value, 4, i);
			out += "-";
			i = printChars(&out, m_value, 2, i);
			out += "-";
			i = printChars(&out, m_value, 2, i);
			out += "-";
			i = printChars(&out, m_value, 2, i);
			out += "-";
			i = printChars(&out, m_value, 6, i);
			return out;
		}
};

}
